

// Name
//===================================
//
// Graphics
//


// DEPENDENCIES
//===================================

    #include "graphics.h"
    #include <SDL2/SDL_image.h>
    #include "unit.h"


// STATIC PROPERTIES
//===================================

    TimerRepeat Graphics::animation_frame_ = TimerRepeat();


// METHODS
//===================================

    Graphics::Graphics()
    {
        adjustMagnification();

        createWindow();
        createRenderer();
    };

    Graphics::~Graphics()
    {
        deletePalette();

        clearTextures();

        for ( int i = 0; i < SpriteSheet::LIMIT; ++i )
        {
            if ( surfaces_[ i ] != nullptr )
            {
                SDL_FreeSurface( surfaces_[ i ] );
            }
        }

        sdl2::SDL_Deleter( renderer_ );
        sdl2::SDL_Deleter( window_ );
        IMG_Quit();
    };

    const std::string Graphics::imgAddress( const std::string& relative_path ) const
    {
        const char* base_path = SDL_GetBasePath();

        if ( base_path == nullptr )
        {
            base_path = SDL_strdup("./");
        }

        const std::string address = base_path + relative_path;

        return address;
    };

    SDL_Texture* Graphics::loadTexture( SpriteSheet sheet )
    {
        if ( surfaces_[ sheet ] == nullptr )
        {
            SDL_RWops* rwop = SDL_RWFromFile( imgAddress( graphic_addresses_[ (int)sheet ] ).c_str(), "rb" );

            if ( rwop == nullptr )
            {
                SDL_Log( "SDL_RWFromFile failed: %s", SDL_GetError() );
            }

            surfaces_[ sheet ] = IMG_LoadPNG_RW( rwop );

            if ( surfaces_[ sheet ] == nullptr )
            {
                SDL_Log( "SDL_Surface failed: %s", SDL_GetError() );
            }
        }

        if ( palette_ != nullptr )
            palette_->applyPalette( surfaces_[ sheet ] );

        SDL_Texture* texture = SDL_CreateTextureFromSurface( renderer_.get(), surfaces_[ sheet ] );

        //SDL_FreeSurface( surface );

        if ( texture == nullptr )
        {
            SDL_Log( "SDL_CreateTextureFromSurface failed: %s", SDL_GetError() );
        }

        return texture;
    };

    bool Graphics::createRenderer()
    {
        if ( window_ == nullptr )
            return false;

        renderer_.reset( SDL_CreateRenderer( window_.get(), -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC ) );

        if ( renderer_ == nullptr )
        {
            SDL_Log( "SDL_CreateRenderer failed: %s", SDL_GetError() );
            return false;
        }

        return true;
    };

    bool Graphics::createWindow()
    {
        window_.reset( SDL_CreateWindow
        (
            "Boskeopolis Land",
            0,
            0,
            magnified( Unit::WINDOW_WIDTH_PIXELS ),
            magnified( Unit::WINDOW_HEIGHT_PIXELS ),
            //SDL_WINDOW_RESIZABLE
            SDL_WINDOW_FULLSCREEN_DESKTOP
        ) );

        if ( window_ == nullptr )
        {
            SDL_Log( "SDL_WindowCreation failed: %s", SDL_GetError() );
            return false;
        }

        return true;
    };

    void Graphics::clearTextures()
    {/*
        for ( auto& i : textures_ )
        {
            if ( i != nullptr )
            {
                //i.reset();
                sdl2::SDL_Deleter( i );
            }
        }*/

        for ( int i = 0; i < SpriteSheet::LIMIT; ++i )
        {
            if ( textures_[ i ] != nullptr )
            {
                //sdl2::SDL_Deleter( textures_[ i ] );
                SDL_DestroyTexture( textures_[ i ] );
                textures_[ i ] = nullptr;
            }
        }

    };

    void Graphics::colorCanvas()
    {
        int r = 255;
        int g = 255;
        int b = 255;

        if ( palette_ != nullptr )
        {
            r = palette_->bgR();
            g = palette_->bgG();
            b = palette_->bgB();
        }

        SDL_SetRenderDrawColor( renderer_.get(), r, g, b, 255 );
        SDL_RenderFillRect( renderer_.get(), &screen_ );
    };

    void Graphics::screenBorders()
    {
        SDL_SetRenderDrawColor( renderer_.get(), 0, 0, 0, 255 );
        SDL_RenderFillRect( renderer_.get(), &top_bar_ );
        SDL_RenderFillRect( renderer_.get(), &bottom_bar_ );
        SDL_RenderFillRect( renderer_.get(), &left_bar_ );
        SDL_RenderFillRect( renderer_.get(), &right_bar_ );
    };

    void Graphics::clearScreen()
    {
        SDL_SetRenderDrawColor( renderer_.get(), 0, 0, 0, 255 );
        SDL_RenderClear( renderer_.get() );
    };

    void Graphics::presentScreen()
    {
        SDL_RenderPresent( renderer_.get() );
    };

    void Graphics::renderObject( SpriteSheet sheet, SDL_Rect source, SDL_Rect dest, SDL_RendererFlip flip, double rotation )
    {
        if ( sheet < SpriteSheet::LIMIT )
        {
            if ( textures_[ sheet ] == nullptr )
            {
                textures_[ sheet ] = loadTexture( sheet );
            }

            if ( textures_[ sheet ] != nullptr )
            {
                dest = sourceRelativeToScreen( dest );
                if ( SDL_RenderCopyEx( renderer_.get(), textures_[sheet], &source, &dest, rotation, 0, flip ) != 0 )
                    printf( "Render failure: %s\n", SDL_GetError() );
            }
        }
    };

    int Graphics::magnified( int n ) const
    {
        return n * magnification_;
    };


    void Graphics::adjustMagnification()
    {
        SDL_Rect r;
        if ( SDL_GetDisplayBounds( 0, &r ) != 0 )
        {
            SDL_Log( "SDL_GetDisplayBounds failed: %s", SDL_GetError() );
            magnification_ = 1;
        }
        else
        {
            int monitor_width = r.w;
            int monitor_height = r.h;

            double screen_aspect_ratio = (double)Unit::WINDOW_WIDTH_PIXELS / (double)Unit::WINDOW_HEIGHT_PIXELS;
            double monitor_aspect_ratio = (double)monitor_width / (double)monitor_height;

            if ( monitor_aspect_ratio > screen_aspect_ratio )
            {
                magnification_ = floor( monitor_height / (double)Unit::WINDOW_HEIGHT_PIXELS );
            }
            else
            {
                magnification_ = floor( monitor_width / (double)Unit::WINDOW_WIDTH_PIXELS );
            }

            left_edge_ = ( monitor_width - magnified( Unit::WINDOW_WIDTH_PIXELS ) ) / 2;
            top_edge_ = ( monitor_height - magnified( Unit::WINDOW_HEIGHT_PIXELS ) ) / 2;

            adjustScreen();

            top_bar_ = { 0, 0, monitor_width, top_edge_ };
            bottom_bar_ = { 0, screen_.y + screen_.h, monitor_width, monitor_height - ( screen_.y + screen_.h ) };
            left_bar_ = { 0, 0, left_edge_, monitor_height };
            right_bar_ = { screen_.x + screen_.w, 0, monitor_width - ( screen_.x + screen_.w ), monitor_height };
        }
    };

    void Graphics::adjustScreen()
    {
        screen_.x = left_edge_;
        screen_.y = top_edge_;
        screen_.w = magnified( Unit::WINDOW_WIDTH_PIXELS );
        screen_.h = magnified( Unit::WINDOW_HEIGHT_PIXELS );
    };

    SDL_Rect Graphics::sourceRelativeToScreen( SDL_Rect source ) const
    {
        return
        {
            magnified( source.x ) + left_edge_,
            source.y = magnified( source.y ) + top_edge_,
            source.w = magnified( source.w ),
            source.h = magnified( source.h ),
        };
    };

    void Graphics::renderRect( SDL_Rect box, int color, int alpha )
    {
        int r = 0;
        int g = 0;
        int b = 0;

        if ( palette_ != nullptr )
        {
            SDL_Color c = palette_->getColor( color );

            r = c.r;
            g = c.g;
            b = c.b;
        }

        SDL_Rect box_relative = sourceRelativeToScreen( box );
        SDL_SetRenderDrawColor( renderer_.get(), r, g, b, alpha );
        SDL_RenderFillRect( renderer_.get(), &box_relative );
    };

    void Graphics::newPalette( Palette::PaletteSet palette )
    {
        deletePalette();

        palette_ = new Palette( palette );
        reloadTextures();
    };

    void Graphics::reloadTextures()
    {
        clearTextures();/*
        for ( int i = 0; i < SpriteSheet::LIMIT; ++i )
        {
            textures_[ i ].reset( loadTexture( graphic_addresses_[ i ] ) );
        }*/
    };

    void Graphics::deletePalette()
    {
        if ( palette_ != nullptr )
        {
            delete palette_;
            palette_ = nullptr;
        }
    };

    void Graphics::update() { animation_frame_.update(); };
    bool Graphics::nextFrame() { return animation_frame_.hit(); };
