
// Name
//===================================
//
// Sprite
//


// DEPENDENCIES
//===================================

    #include "block.h"
    #include "camera.h"
    #include "collision.h"
    #include "inventory.h"
    #include "map.h"
    #include "sprite.h"
    #include "sprite_graphics.h"
    #include "input.h"
    #include "unit.h"


// STATIC PROPERTIES
//===================================

    double Sprite::traction_ = TRACTION_NORMAL;
    int Sprite::resistance_x_ = RESISTANCE_X_NORMAL;


// METHODS
//===================================

    Sprite::Sprite
    (
        SpriteGraphics* graphics,
        int x,
        int y,
        int width,
        int height,
        SpriteType type,
        int start_speed,
        int top_speed,
        int jump_start_speed,
        int jump_top_speed
    )
    :
        Object( x, y, width, height ),
        graphics_ ( graphics ),
        type_ ( type ),
        jump_start_speed_ ( jump_start_speed ),
        jump_top_speed_normal_ ( jump_top_speed ),
        jump_top_speed_ ( jump_top_speed ),
        start_speed_walk_ ( start_speed ),
        top_speed_walk_ ( top_speed ),
        start_speed_ ( start_speed ),
        top_speed_ ( top_speed ),
        start_speed_run_ ( start_speed*2 ),
        top_speed_run_ ( top_speed*2 )
    {};

    Sprite::~Sprite()
    {
        if ( graphics_ != nullptr )
        {
            delete graphics_;
            graphics_ = nullptr;
        }
    };

    bool Sprite::fellInBottomlessPit( Map& lvmap ) const
    {
        return hit_box_.y > Unit::PixelsToSubPixels( lvmap.heightPixels() );
    };

    void Sprite::testInvincibility()
    {
        if ( invincible_ )
        {
            invincibility_timer_.update();

            if ( invincibility_timer_.done() )
            {
                invincible_ = false;
            }
            else if ( invincibility_timer_.counter() % 4 == 1 )
            {
                if ( graphics_ != nullptr )
                    graphics_->visible_ = false;
            }
        }
    };

    void Sprite::update( Input& input, Camera& camera, Map& lvmap, Game& game, EventSystem& events )
    {
        customUpdate( input, camera, lvmap, game, events );

        position();

        if ( fellInBottomlessPit( lvmap ) )
            kill();

        if ( graphics_ != nullptr )
            graphics_->visible_ = true;

        testInvincibility();

        if ( graphics_ != nullptr )
        {
            graphics_->update( *this );
        }
    };

    void Sprite::render( Graphics& graphics, Camera& camera, bool priority )
    {
        if ( graphics_ != nullptr )
        {
            graphics_->render( graphics, Unit::SubPixelsToPixels( hit_box_ ), &camera, priority );
        }
    };

    void Sprite::moveLeft()
    {
        acceleration_x_ = -start_speed_;
        direction_x_ = Direction::Horizontal::LEFT;
    };

    void Sprite::moveRight()
    {
        acceleration_x_ = start_speed_;
        direction_x_ = Direction::Horizontal::RIGHT;
    };

    void Sprite::stop()
    {
        acceleration_x_ = 0;
    };

    void Sprite::stopDucking()
    {
        acceleration_x_ = 0;
    };

    void Sprite::positionX()
    {
        if ( isJumping() )
        {
            acceleration_x_ *= .75;
        }

        if ( !onGround() )
            top_speed_ = top_speed_walk_ * .95;

        if ( isDucking() )
        {
            acceleration_x_ /= 3;
        }

        vx_ += acceleration_x_;

        if ( acceleration_x_ == 0 )
            vx_ /= traction_;

        if ( vx_ > top_speed_ )
            vx_ = top_speed_;
        if ( vx_ < -top_speed_ )
            vx_ = -top_speed_;

        vx_ += resistance_x_;

        hit_box_.x += vx_;

        top_speed_ = top_speed_walk_;
        start_speed_ = start_speed_walk_;

        is_running_ = false;
    };

    void Sprite::positionY()
    {
        if ( isDucking() )
        {
            jump_top_speed_ = JUMP_DUCK_TOP_SPEED;
        }

        if ( abs( top_speed_ ) > top_speed_walk_ )
            jump_top_speed_ *= 1.5;

        if ( !onGround() && !isJumping() && !onLadder() )
        {
            acceleration_y_ = gravity_start_speed_;
        }

        vy_ += acceleration_y_;

        if ( vy_ > gravity_top_speed_ )
            vy_ = gravity_top_speed_;

        if ( vy_ < -jump_top_speed_ )
        {
            vy_ = -jump_top_speed_;
            jump_start_ = false;
        }

        hit_box_.y += vy_;

        gravity_start_speed_ = GRAVITY_START_SPEED;
        gravity_top_speed_ = GRAVITY_TOP_SPEED;
        jump_top_speed_ = jump_top_speed_normal_;
    };

    void Sprite::position()
    {
        x_prev_ = hit_box_.x;
        y_prev_ = hit_box_.y;

        if ( !isDead() )
            positionX();

        if ( !flying_ )
            positionY();


        if ( acceleration_x_ != 0 || acceleration_y_ != 0 )
            is_moving_ = true;
        else
            is_moving_ = false;

        acceleration_y_ = 0;

        if ( !isJumping() )
            jump_end_ = true;

        is_jumping_ = false;
        on_ground_prev_ = on_ground_;
        on_ground_ = false;
        collide_top_prev_ = collide_top_;
        collide_bottom_prev_ = collide_bottom_;
        collide_top_ = false;
        collide_bottom_ = false;
        collide_left_ = false;
        collide_right_ = false;
        touching_ladder_prev_ = touching_ladder_;
        touching_ladder_ = false;
        is_sliding_ = false;

        if ( onLadder() )
        {
            vy_ = 0;
        }
    };

    void Sprite::boundaries( Camera& camera )
    {
        if ( hit_box_.x < camera.x() )
        {
            collideStopXLeft( abs ( camera.x() - hit_box_.x ) );
        }
        if ( hit_box_.x + hit_box_.w > Unit::PixelsToSubPixels( camera.right() ) )
        {
            collideStopXRight( ( hit_box_.x + hit_box_.w ) - Unit::PixelsToSubPixels( camera.right() ) );
        }
    };

    void Sprite::collideStopXLeft( int overlap )
    {
        stop();

        if ( vx_ < 0 )
        {
            hit_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
        }
    };

    void Sprite::collideStopXRight( int overlap )
    {
        stop();

        if ( vx_ > 0 )
        {
            hit_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
        }
    };

    void Sprite::collideStopYBottom( int overlap )
    {
        if ( vy_ > 0 )
        {
            acceleration_y_ = 0;

            if ( !onGround() )
            {
                hit_box_.y -= overlap;
            }

            on_ground_ = true;
        }

        if ( onLadder() )
        {
            hit_box_.y -= LADDER_SPEED;
        }
    };

    void Sprite::collideStopYTop( int overlap )
    {
        acceleration_y_ = 0;
        hit_box_.y += vy_;
        vy_ = 0;
        hit_box_.y += overlap;
        jump_end_ = true;
    };

    bool Sprite::onGround() const
    {
        return on_ground_;
    };

    bool Sprite::isJumping() const
    {
        return is_jumping_;
    };

    void Sprite::jump()
    {
        if ( !onGround() || !jump_lock_ )
        {
            if ( onGround() && !jump_lock_ )
            {
                startJump();
            }

            if ( jump_start_ && !jump_end_ )
            {
                is_jumping_ = true;
            }

            if ( isJumping() )
            {
                acceleration_y_ = -jump_start_speed_;
            }
        }
    };

    void Sprite::startJump()
    {
        jump_start_ = true;
        jump_end_ = false;
        vy_ = 0;
        on_ground_ = false;
    };

    void Sprite::slowFall()
    {
        gravity_start_speed_ = GRAVITY_START_SPEED * .75;
        gravity_top_speed_ = GRAVITY_TOP_SPEED * .85;
    };

    bool Sprite::isRunning() const
    {
        return is_running_;
    };

    void Sprite::run()
    {
        is_running_ = true;
        top_speed_ = top_speed_run_;
        start_speed_ = start_speed_run_;
    };

    SDL_Rect Sprite::adjustBoundBoxes()
    {

        int h = heightPixels() + 1;
        int y = yPixels() - 1;

        if ( isDucking() )
        {
            h += 4;
            y -= 4;
        }

        return
        {
            xPixels() - 1,
            y,
            widthPixels() + 2,
            h
        };
    };

    void Sprite::interact( Sprite& them )
    {
        Collision my_collision = collide( &them );
        Collision their_collision = them.collide( this );

        customInteract( my_collision, their_collision, them );
    };

    bool Sprite::canJump() const
    {
        return can_jump_;
    };

    bool Sprite::isDucking() const
    {
        return is_ducking_;
    };

    void Sprite::duck()
    {
        // Can continue ducking while in air, but only start duck on ground.
        if ( isDucking() || onGround() )
        {
            // Hacky way to make player warp to the right position after height changes.
            if ( !isDucking() )
            {
                hit_box_.y += 10000;
                graphics_->y_adjustment_ = -5;
                graphics_->h_adjustment_ = 6;
            }

            is_ducking_ = true;

            hit_box_.h = Unit::PixelsToSubPixels( original_height_ - 11 );
        }
    };

    void Sprite::unduck()
    {
        // Hacky way to keep player from falling through ground after gaining height from unducking.
        if ( isDucking() )
        {
            hit_box_.y -= 10000;
            graphics_->y_adjustment_ = -1;
            graphics_->h_adjustment_ = 2;
        }

        is_ducking_ = false;
        hit_box_.h = Unit::PixelsToSubPixels( original_height_ - 1 );
    };

    Direction::Horizontal Sprite::directionX() const
    {
        return direction_x_;
    };

    bool Sprite::onGroundPrev() const
    {
        return on_ground_prev_;
    };

    bool Sprite::isMoving() const
    {
        return is_moving_;
    };

    void Sprite::stopY()
    {
        acceleration_y_ = 0;
        vy_ = 0;
    };

    bool Sprite::onLadder() const
    {
        return on_ladder_;
    };

    bool Sprite::onLadderPrev() const
    {
        return on_ladder_prev_;
    };

    bool Sprite::touchingLadder() const
    {
        return touching_ladder_;
    };

    bool Sprite::touchingLadderPrev() const
    {
        return touching_ladder_prev_;
    };

    void Sprite::touchLadder()
    {
        touching_ladder_ = true;
    };

    void Sprite::grabLadder()
    {
        if ( !isDucking() )
            on_ladder_ = true;
    };

    void Sprite::releaseLadder()
    {
        on_ladder_ = false;
    };

    void Sprite::moveUp()
    {
        if ( onLadder() )
        {
            acceleration_y_ = -LADDER_SPEED;
        }
        else if ( onGround() )
        {
            lookUp();
        }
    };

    void Sprite::moveDown()
    {
        if ( onLadder() )
        {
            acceleration_y_ = LADDER_SPEED;
        }
    };

    void Sprite::lookUp()
    {
        looking_up_ = true;
    };

    bool Sprite::lookingUp() const
    {
        return looking_up_;
    };

    void Sprite::stopLookingUp()
    {
        looking_up_ = false;
    };

    void Sprite::kill()
    {
        dead_ = true;
    };

    bool Sprite::isDead() const
    {
        return dead_;
    };

    void Sprite::setPosition( int x, int y )
    {
        hit_box_.x = Unit::PixelsToSubPixels( x );
        hit_box_.y = Unit::PixelsToSubPixels( y );
    };

    bool Sprite::hasType( Sprite::SpriteType type ) const
    {
        if ( type_ == type )
            return true;
        else return false;
    };

    void Sprite::hurt( int amount )
    {
        if ( !invincible_ )
        {
            hp_ -= amount;

            if ( hp_ <= 0 )
            {
                kill();
            }
            else
            {
                invincible_ = true;
                invincibility_timer_.start();
            }
        }
    };

    void Sprite::heal( int amount )
    {
        hp_ += amount;

        if ( hp_ > max_hp_ )
            hp_ = max_hp_;
    };

    int Sprite::hp() const
    {
        return hp_;
    };


    void Sprite::collideStopAny( Collision& collision )
    {
        if ( !collide_bottom_prev_ || !collide_top_prev_ )
        {
            if ( collision.collideLeft() )
            {
                collide_left_ = true;

                collideStopXLeft( collision.overlapXLeft() );
            }

            if ( collision.collideRight() )
            {
                collide_right_ = true;

                collideStopXRight( collision.overlapXRight() );
            }
        }

        if ( collision.collideBottom() )
        {
            collideStopYBottom( collision.overlapYBottom() );
            collide_top_ = true;
        }

        if ( collision.collideTop() )
        {
            collide_bottom_ = true;

            if ( !collide_top_prev_ )
            {
                collideStopYTop( collision.overlapYTop() );
            }
        }
    };


    //void Sprite::customUpdate( Input& input, Camera& camera, Map& lvmap, Game& game, EventSystem& events ) {};
    //void Sprite::customInteract( Collision& my_collision, Collision& their_collision, Sprite& them ) {};

    bool Sprite::canInteract() const
    {
        return interact_;
    };

    bool Sprite::collidedRight() const
    {
        return collide_right_;
    };

    bool Sprite::collidedLeft() const
    {
        return collide_left_;
    };

    void Sprite::reset()
    {
        hit_box_.x = original_x_;
        hit_box_.y = original_y_;
    };

    void Sprite::slideLeft()
    {
        is_sliding_ = true;
        moveLeft();
    };

    void Sprite::slideRight()
    {
        is_sliding_ = true;
        moveRight();
    };
