
// Name
//===================================
//
// Sprite
//

#ifndef SPRITE_H
#define SPRITE_H


// FORWARD DECLARATIONS
//===================================

    class Block;
    class Camera;
    class Collision;
    class EventSystem;
    class Game;
    class GameState;
    class Input;
    class Inventory;
    class Level;
    class Map;
    class SpriteGraphics;


// DEPENDENCIES
//===================================

    #include "direction.h"
    #include "graphics.h"
    #include "object.h"
    #include <SDL2/SDL.h>
    #include "timers/timer_simple.h"
    #include "unit.h"


// CLASS
//===================================

    class Sprite : public Object
    {
        public:
            enum SpriteType
            {
                HERO,
                ENEMY,
                NOTYPE
            };

            Sprite
            (
                SpriteGraphics* graphics = nullptr,
                int x = 0,
                int y = 0,
                int width = 16,
                int height = 16,
                SpriteType type = SpriteType::NOTYPE,
                int start_speed = 160,
                int top_speed = 2000,
                int jump_start_speed = 1000,
                int jump_top_speed = 6000
            );
            virtual ~Sprite();

            static constexpr double TRACTION_NORMAL = 1.1;
            static constexpr double TRACTION_ICY = 1.025;
            static const int RESISTANCE_X_NORMAL = 0;
            static const int RESISTANCE_X_WINDY = 100;

            static double traction_;
            static int resistance_x_;

            int vx_ = 0;
            int vy_ = 0;

            Direction::Horizontal direction_x_ = Direction::Horizontal::RIGHT;
            bool touching_ladder_ = false;
            bool on_ground_ = false;
            bool is_sliding_ = false;

            void update( Input& input, Camera& camera, Map& lvmap, Game& game, EventSystem& events );
            void render( Graphics& graphics, Camera& camera, bool priority = false );
            void interact( Sprite& them );

            virtual void customUpdate( Input& input, Camera& camera, Map& lvmap, Game& game, EventSystem& events ) = 0;
            virtual void customInteract( Collision& my_collision, Collision& their_collision, Sprite& them ) = 0;
            virtual void reset();

            bool onGround() const;
            bool onGroundPrev() const;
            bool canJump() const;
            bool isJumping() const;
            bool isRunning() const;
            bool isDucking() const;
            bool isMoving() const;
            bool onLadder() const;
            bool onLadderPrev() const;
            bool touchingLadder() const;
            bool touchingLadderPrev() const;
            bool lookingUp() const;
            bool isDead() const;
            bool collidedLeft() const;
            bool collidedRight() const;
            bool canInteract() const;
            bool hasType( SpriteType type ) const;
            int hp() const;

            void moveLeft();
            void moveRight();
            void moveUp();
            void moveDown();
            void slideLeft();
            void slideRight();
            void stop();
            void stopY();
            void collideStopAny( Collision& collision );
            void collideStopXLeft( int overlap );
            void collideStopXRight( int overlap );
            void collideStopYBottom( int overlap );
            void collideStopYTop( int overlap );
            void jump();
            void run();
            void duck();
            void stopDucking();
            void unduck();
            void slowFall();
            void lookUp();
            void stopLookingUp();
            void kill();
            void boundaries( Camera& camera );
            void hurt( int amount = 1 );
            void heal( int amount = 1 );
            void touchLadder();
            void grabLadder();
            void releaseLadder();

            void setPosition( int x, int y );
            Direction::Horizontal directionX() const;


        protected:
            static const int GRAVITY_START_SPEED = 500;
            static const int GRAVITY_TOP_SPEED = 4000;
            static const int JUMP_DUCK_TOP_SPEED = 5500;
            const double BOUNCE = .2;
            static const int LADDER_SPEED = 1000;

            TimerSimple invincibility_timer_ = TimerSimple( 48, false );
            const SpriteType type_;
            SpriteGraphics* graphics_;

            int start_speed_walk_;
            int top_speed_walk_;
            int start_speed_run_;
            int top_speed_run_;
            int jump_start_speed_;
            int jump_top_speed_normal_;

            int acceleration_x_ = 0;
            int acceleration_y_ = 0;
            int max_hp_ = 3;
            int hp_ = max_hp_;
            int gravity_start_speed_ = GRAVITY_START_SPEED;
            int gravity_top_speed_ = GRAVITY_TOP_SPEED;
            int start_speed_;
            int top_speed_;
            int jump_top_speed_;

            bool jump_lock_ = false;
            bool collide_top_ = false;
            bool collide_bottom_ = false;
            bool collide_top_prev_ = false;
            bool collide_bottom_prev_ = false;
            bool collide_left_ = false;
            bool collide_right_ = false;
            bool collide_left_prev_ = false;
            bool collide_right_prev_ = false;
            bool touching_ladder_prev_ = false;
            bool flying_ = false;
            bool interact_ = true;
            bool invincible_ = false;
            bool is_running_ = false;
            bool on_ground_prev_ = false;
            bool can_jump_ = false;
            bool is_jumping_ = false;
            bool jump_start_ = false;
            bool jump_end_ = false;
            bool is_ducking_ = false;
            bool is_moving_ = false;
            bool on_ladder_ = false;
            bool on_ladder_prev_ = false;
            bool looking_up_ = false;
            bool dead_ = false;

            void position();
            void positionX();
            void positionY();
            void testInvincibility();
            bool fellInBottomlessPit( Map& lvmap ) const;
            SDL_Rect adjustBoundBoxes();
            void startJump();
    };

#endif // SPRITE_H

